/*
 * Copyright (C) 2002  Pekka Enberg <Pekka.Enberg@cs.helsinki.fi>
 *
 * Distributed under the terms of the GNU General Public License
 * version 2 or later.
 */
#include <assert.h>
#include <curses.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>

#define ERROR_MSG_BUFFER 1024

void
error (const char * format, ...)
{
  va_list args;
  char buffer[ERROR_MSG_BUFFER];

  endwin ();
  va_start (args, format);
  vsnprintf (buffer, ERROR_MSG_BUFFER, format, args);
  va_end (args);
  printf ("error: %s\n", buffer);
  exit (EXIT_FAILURE);
}

/*
 * Edit buffer attribute -> curses color pair lookup table.
 */
char attr_to_color_pair[8][8] = {
  {  0,  1,  2,  3,  4,  5,  6,  7 },
  {  8,  9, 10, 11, 12, 13, 14, 15 },
  { 16, 17, 18, 19, 20, 21, 22, 23 },
  { 24, 25, 26, 27, 28, 29, 30, 31 },
  { 32, 33, 34, 35, 36, 37, 38, 39 },
  { 40, 41, 42, 43, 44, 45, 46, 47 },
  { 48, 49, 50, 51, 52, 53, 54, 55 },
  { 56, 57, 58, 59, 60, 61, 62, 63 }
};

void
init_color_pairs ()
{
  if (COLOR_PAIRS < 64)
    {
      error("Cannot initialize color pairs. Only %i supported.", COLOR_PAIRS);
    }

  /* All color pairs (foreground/background combinations) should be
     initialized here.  */
  init_pair ( 0, COLOR_BLACK,   COLOR_BLACK);
  init_pair ( 1, COLOR_RED,     COLOR_BLACK);
  init_pair ( 2, COLOR_GREEN,   COLOR_BLACK);
  init_pair ( 3, COLOR_YELLOW,  COLOR_BLACK);
  init_pair ( 4, COLOR_BLUE,    COLOR_BLACK);
  init_pair ( 5, COLOR_MAGENTA, COLOR_BLACK);
  init_pair ( 6, COLOR_CYAN,    COLOR_BLACK);
  init_pair ( 7, COLOR_WHITE,   COLOR_BLACK);

  init_pair ( 8, COLOR_BLACK,   COLOR_RED);
  init_pair ( 9, COLOR_RED,     COLOR_RED);
  init_pair (10, COLOR_GREEN,   COLOR_RED);
  init_pair (11, COLOR_YELLOW,  COLOR_RED);
  init_pair (12, COLOR_BLUE,    COLOR_RED);
  init_pair (13, COLOR_MAGENTA, COLOR_RED);
  init_pair (14, COLOR_CYAN,    COLOR_RED);
  init_pair (15, COLOR_WHITE,   COLOR_RED);

  init_pair (16, COLOR_BLACK,   COLOR_GREEN);
  init_pair (17, COLOR_RED,     COLOR_GREEN);
  init_pair (18, COLOR_GREEN,   COLOR_GREEN);
  init_pair (19, COLOR_YELLOW,  COLOR_GREEN);
  init_pair (20, COLOR_BLUE,    COLOR_GREEN);
  init_pair (21, COLOR_MAGENTA, COLOR_GREEN);
  init_pair (22, COLOR_CYAN,    COLOR_GREEN);
  init_pair (23, COLOR_WHITE,   COLOR_GREEN);

  init_pair (24, COLOR_BLACK,   COLOR_YELLOW);
  init_pair (25, COLOR_RED,     COLOR_YELLOW);
  init_pair (26, COLOR_GREEN,   COLOR_YELLOW);
  init_pair (27, COLOR_YELLOW,  COLOR_YELLOW);
  init_pair (28, COLOR_BLUE,    COLOR_YELLOW);
  init_pair (29, COLOR_MAGENTA, COLOR_YELLOW);
  init_pair (30, COLOR_CYAN,    COLOR_YELLOW);
  init_pair (31, COLOR_WHITE,   COLOR_YELLOW);

  init_pair (32, COLOR_BLACK,   COLOR_BLUE);
  init_pair (33, COLOR_RED,     COLOR_BLUE);
  init_pair (34, COLOR_GREEN,   COLOR_BLUE);
  init_pair (35, COLOR_YELLOW,  COLOR_BLUE);
  init_pair (36, COLOR_BLUE,    COLOR_BLUE);
  init_pair (37, COLOR_MAGENTA, COLOR_BLUE);
  init_pair (38, COLOR_CYAN,    COLOR_BLUE);
  init_pair (39, COLOR_WHITE,   COLOR_BLUE);

  init_pair (40, COLOR_BLACK,   COLOR_MAGENTA);
  init_pair (41, COLOR_RED,     COLOR_MAGENTA);
  init_pair (42, COLOR_GREEN,   COLOR_MAGENTA);
  init_pair (43, COLOR_YELLOW,  COLOR_MAGENTA);
  init_pair (44, COLOR_BLUE,    COLOR_MAGENTA);
  init_pair (45, COLOR_MAGENTA, COLOR_MAGENTA);
  init_pair (46, COLOR_CYAN,    COLOR_MAGENTA);
  init_pair (47, COLOR_WHITE,   COLOR_MAGENTA);

  init_pair (48, COLOR_BLACK,   COLOR_CYAN);
  init_pair (49, COLOR_RED,     COLOR_CYAN);
  init_pair (50, COLOR_GREEN,   COLOR_CYAN);
  init_pair (51, COLOR_YELLOW,  COLOR_CYAN);
  init_pair (52, COLOR_BLUE,    COLOR_CYAN);
  init_pair (53, COLOR_MAGENTA, COLOR_CYAN);
  init_pair (54, COLOR_CYAN,    COLOR_CYAN);
  init_pair (55, COLOR_WHITE,   COLOR_CYAN);

  init_pair (56, COLOR_BLACK,   COLOR_WHITE);
  init_pair (57, COLOR_RED,     COLOR_WHITE);
  init_pair (58, COLOR_GREEN,   COLOR_WHITE);
  init_pair (59, COLOR_YELLOW,  COLOR_WHITE);
  init_pair (60, COLOR_BLUE,    COLOR_WHITE);
  init_pair (61, COLOR_MAGENTA, COLOR_WHITE);
  init_pair (62, COLOR_CYAN,    COLOR_WHITE);
  init_pair (63, COLOR_WHITE,   COLOR_WHITE);
}

void
set_attribute (int attribute)
{
  int fg_color = (attribute & 0x0F);
  int bg_color = (attribute & 0xF0) >> 4;

  /* We need to turn on the BOLD attribute to get colors > 7.  */
  if (fg_color > 7)
    {
      attron (A_BOLD);
      fg_color -= 7;
    }
  else
    {
      attroff (A_BOLD);
    }
  attron (COLOR_PAIR (attr_to_color_pair[bg_color][fg_color]));
}

#define COLOR_ATTR(fg, bg) (bg * 0x10 + fg)

/*
 * If you use any of the colors, please define a constant here instead
 * of hardcoding the value into the code.
 */
#define RED_ON_BLACK 1

void
print_status (int screen_height, size_t x, size_t y,
	      int fg_color, int bg_color)
{
  move (screen_height - 1, 0);
  clrtoeol ();

  attron (COLOR_PAIR(RED_ON_BLACK));
  mvprintw (screen_height - 1, 1, "(%2i, %2i)", x, y);
  attroff (COLOR_PAIR(RED_ON_BLACK));

  move (screen_height - 1, 13);
  set_attribute (COLOR_ATTR (fg_color, bg_color));
  printw ("Color");
}

struct edit_buffer {
  size_t height;
  size_t width;
  int * buffer;
};

void
edit_buffer_put (struct edit_buffer * buffer, size_t x, size_t y, int value)
{
  assert (x < buffer->width ); assert (x >= 0);
  assert (y < buffer->height); assert (y >= 0);

  buffer->buffer[y * buffer->width + x] = value;
}

int
edit_buffer_get (struct edit_buffer * buffer, size_t x, size_t y)
{
  assert (x < buffer->width ); assert(x >= 0);
  assert (y < buffer->height); assert(y >= 0);

  return buffer->buffer[y * buffer->width + x];
}

void
edit_buffer_clear (struct edit_buffer * buffer)
{
  int x, y;

  for (y = 0; y < buffer->height; y++)
    for (x = 0; x < buffer->width; x++)
      edit_buffer_put (buffer, x, y, 0x0720);
}

void
edit_buffer_draw_to_screen (struct edit_buffer * buffer, size_t start_x,
			    size_t start_y, size_t screen_width,
			    size_t screen_height)
{
  assert (start_x + screen_width  <= buffer->width );
  assert (start_y + screen_height <= buffer->height);
  assert (start_x >= 0);
  assert (start_y >= 0);

  int x, y;

  for (y = 0; y < screen_height; y++)
    for (x = 0; x < screen_width; x++)
      {
	int attribute =
	  (edit_buffer_get (buffer, start_x + x, start_y + y) & 0xFF00) >> 8;
	int character =
	  edit_buffer_get (buffer, start_x + x, start_y + y) & 0xFF;

	set_attribute (attribute);
	mvprintw (y, x, "%c", character);
      }
}

void
edit_buffer_init (struct edit_buffer * buffer, size_t width, size_t height)
{
  buffer->buffer = malloc(height * width * sizeof(int));
  if (buffer->buffer == NULL)
    {
      error("Could not allocate memory for edit buffer.");
    }
  buffer->height = height;
  buffer->width = width;
}

void
edit_buffer_release (struct edit_buffer * buffer)
{
  if (buffer->buffer != NULL)
    free(buffer->buffer);

  buffer->buffer = NULL;
  buffer->width = 0;
  buffer->height = 0;
}

void
edit_loop ()
{
  bool quit = false;

  int fg_color = 0x07;
  int bg_color = 0x00;

  int edit_buffer_start_x = 0;
  int edit_buffer_start_y = 0;
  int screen_x = 0;
  int screen_y = 0;

  int screen_height, screen_width;

  getmaxyx (stdscr, screen_height, screen_width);
  if (screen_height == -1 || screen_width == -1)
    {
      error("Could not get screen dimensions from curses.");
    }

  /* Leave a line for the status bar.  */
  screen_height--;

  struct edit_buffer buffer;

  edit_buffer_init (&buffer, 160, 100);
  edit_buffer_clear (&buffer);

  while (!quit)
    {
      edit_buffer_draw_to_screen (&buffer,
				  edit_buffer_start_x,
				  edit_buffer_start_y,
				  screen_width,
				  screen_height);

      print_status (screen_height + 1,
		    edit_buffer_start_x + screen_x + 1,
		    edit_buffer_start_y + screen_y + 1,
		    fg_color,
		    bg_color);
      
      move (screen_y, screen_x);

      int ch = getch ();
      if (ch == ERR)
	{
	  error ("getch() returned ERR");
	}
      switch (ch)
	{
	case KEY_UP:
	  if (screen_y > 0)
	    {
	      screen_y--;
	    }
	  else if (edit_buffer_start_y > 0)
	    {
	      edit_buffer_start_y--;
	    }
	  break;

	case KEY_DOWN:
	  if (screen_y < (screen_height - 1))
	    {
	      screen_y++;
	    }
	  else if (edit_buffer_start_y + screen_height < buffer.height)
	    {
	      edit_buffer_start_y++;
	    }
	  break;

	case KEY_LEFT:
	  if (screen_x > 0)
	    screen_x--;
	  else if (edit_buffer_start_x > 0)
	    edit_buffer_start_x--;
	  break;

	case KEY_RIGHT:
	  if (screen_x < (screen_width - 1))
	    screen_x++;
	  else if (edit_buffer_start_x < (buffer.width - screen_width))
	    edit_buffer_start_x++;
	  break;

	case '+':
	  if (bg_color < 7)
	    bg_color++;
	  break;

	case '-':
	  if (bg_color > 0)
	    bg_color--;
	  break;

	case '*':
	  if (fg_color < 14)
	    fg_color++;
	  break;

	case '/':
	  if (fg_color > 0)
	    fg_color--;
	  break;

	case KEY_END:
	  screen_x = (screen_width - 1);
	  edit_buffer_start_x = buffer.width - screen_width;
	  break;

	case KEY_HOME:
	  screen_x = 0;
	  edit_buffer_start_x = 0;
	  break;

	case KEY_PPAGE:
	  edit_buffer_start_y -= screen_height;
	  if (edit_buffer_start_y < 0)
	    {
	      edit_buffer_start_y = 0;
	      screen_y = 0;
	    }
	  break;

	case KEY_NPAGE:
	  edit_buffer_start_y += screen_height;
	  if (edit_buffer_start_y > buffer.height - screen_height)
	    {
	      edit_buffer_start_y = buffer.height - screen_height;
	      screen_y = (screen_height - 1);
	    }
	  break;

	case KEY_F(10):
	  quit = true;
	  break;

	default:
	  edit_buffer_put (&buffer,
			   edit_buffer_start_x + screen_x,
			   edit_buffer_start_y + screen_y,
			   (COLOR_ATTR(fg_color, bg_color) << 8) | ch);
	  if (screen_x < (screen_width - 1))
	    screen_x++;
	  else if (edit_buffer_start_x < (buffer.width - screen_width))
	    edit_buffer_start_x++;
	  break;
	}
    }
  edit_buffer_release (&buffer);
}

void
init_curses ()
{
  savetty ();
  initscr ();
  nonl ();
  cbreak ();
  noecho ();
  keypad (stdscr, TRUE);

  if (has_colors() == FALSE)
    {
      error("Your terminal doesn't support colors.");
    }
  start_color ();
}

void
release_curses ()
{
  resetty ();
  endwin ();
}

int
main ()
{
  init_curses ();
  init_color_pairs ();
  edit_loop();
  release_curses ();
  printf ("Bye, bye.\n");

  return EXIT_SUCCESS;
}

/*
 * Local Variables:
 * c-file-style: "gnu"
 * End:
 */
